#!/usr/bin/python

##############################################################################
# certs2ldap.py 0.2.2
#
# Alle Zertifikate einer Zertifizierungsstelle auf einem LDAP-Server
# ablegen.
#
# Zu jedem gueltigen Zertifikat in der SSLeay-Zertifikat-DB werden evtl.
# schon bestehende LDAP-Eintraege gesucht und zu diesen das Zertifikat
# als Attribut usersmimecertificate hinzugefgt.
# LDAP-Suche in mehreren Stufen:
# 1. Suche nach Eintraegen mit passenden cn- und mail-Attributen.
# 2. Suche nach Eintraegen mit dem mail-Attribut gesucht.
# 3. Wird ebenfalls kein Eintrag gefunden, so wird ein
#    neuer Eintrag mit den Angaben aus dem Zertifikat angelegt.
# Anmerkung:
# Bestehende Zertifikate werden berschrieben!
##############################################################################

##############################################################################
# Hier einige Variablen zur Konfiguration
##############################################################################

dn_template = 'cn=%s,mail=%s'

# Die Parameter ldap_* koennen auch als Kommandozeilenparameter vorgegeben
# werden. Dies hier sind Vorbelegungen, falls nicht anders angegeben

# Umbenennungstabelle SSLeay-Attributname -> LDAP-Attributname
# Mu angepat werden, falls in SSLeay zustzliche OIDs verwendet werden
ldapattrname = {
                 'C':'c',
                 'ST':'st',
                 'L':'l',
                 'O':'o',
                 'OU':'ou',
                 'CN':'cn',
                 'I':'initials',
                 'uid':'uid',
                 'Email':'mail'
               }

# IP-Adresse oder Name des LDAP-Servers
ldap_host = 'localhost'
# Port-Adresse des LDAP-Servers
ldap_port = 389
# dn zum Binden (Anmelden)
ldap_binddn = ''
# Passwort zu ldap_binddn
ldap_passwd = ''
# ObjectClass fuer neu anzulegende Eintraege
ldap_entryobjectclass = 'person'

import openssl

# Pfadname der SSLeay-Datenbank
openssl.db_filename = '/usr/local/ssl/TestCA/index.txt'
# Verzeichnisname mit SSLeay-Zertfikaten im PEM-Format
openssl_certdir    = '/usr/local/ssl/TestCA/newcerts'
# Pfadname des SSLeay-Programms
openssl_bin         = '/usr/local/ssl/bin/openssl'

##############################################################################
# Ab hier gibt es nix mehr zu konfigurieren
##############################################################################

import sys, os, string, charset, getopt, ldap, openssl

def findoption(options,paramname):
  for i in options:
    if i[0]==paramname:
      return i
  return ()

def PrintErrorMsg(ErrorMsg='',ErrorCode=1,NoExit=0):
  sys.stderr.write('*** Error: %s\n\n' % ErrorMsg)
  if NoExit:
    return
  else:
    sys.exit(ErrorCode)

def PrintUsage(ErrorMsg='',ErrorCode=1):
  script_name = string.split(sys.argv[0],os.sep)[-1]
  sys.stderr.write("""*** %s *** (C) by Michael Stroeder, 1998

usage: %s [-nmr] [-h ldaphost] [-p ldapport] [-D binddn] [-w passwd]
          [-o objectclass] [-d database] [-c certdir] [-s openssl]
	  [filter 1] ... [filter n]

Sends all SSLeay-certicates (PEM-format) matching all filters to a LDAP server.
A filter has the form <attribute>=<value rexexp.> .

Options:

       -n     Create new LDAP entries if no existing LDAP entries
              matching the SSLeay attributes are found. USE WITH CARE!!!
	      Default: false

	      Actually the DN of the new entry has the following format:
	        cn=<Common Name>,mail=<Mail address>
	      This might not match your needs on your LDAP configuration.

	      The attributes of the SSLeay DB entry are used for creating
	      the new LDAP entry. See also option -o for setting the
	      object class. You might also want to configure the Python
	      dictionary ldapattrname in the source to map attribute names
	      for special attributes you are using in certificates.

       -m     Allows to add certificate to LDAP entries for which only the
              the mail address is matched. Otherwise the cn and mail
	      attributes has to match the attributes in the SSLeay DB.
              Default: false

       -r     Allows replacing existing certificate in LDAP entries.
              Default: false

       -D binddn
              Use  binddn  to bind to the X.500 directory. binddn
              should be a string-represented DN as defined in RFC 1779.

       -w passwd
              Use  passwd  as the password for simple authentica
              tion.

       -h ldaphost
              Specify an alternate host on which the ldap  server
              is running.
	      Default: %s

       -p ldapport
              Specify an alternate TCP port where the ldap server
              is listening.
	      Default: %d

       -o ldapobjectclass
              Specify an objectclass which is used for creating new
	      LDAP entries. (See option -c)
	      Default: %s

       -d database
              Specify the pathname of the SSLeay database file.
	      Default: %s

       -c certdir
              Specifies the directory where the certificates are located.
	      Default: %s

       -s openssl
              Specify the pathname of the SSLeay binary.
	      Default: %s

""" % (script_name,script_name,\
       ldap_host,ldap_port,ldap_entryobjectclass,\
       openssl.db_filename,openssl_certdir,openssl_bin))
  if ErrorMsg:
    sys.stderr.write('*** Error: %s\n\n' % ErrorMsg)
  sys.exit(ErrorCode)

##############################################################################
# __main__()
##############################################################################

# Kommandozeile parsen
try:
  options,args=getopt.getopt(sys.argv[1:],'rnmh:p:D:w:o:d:c:s:',[])
except getopt.error:
  PrintUsage('Wrong options!')

# Hostadresse    
try:
  dummy,ldap_host = findoption(options,'-h')
except: pass

# Portnummer
try:
  dummy,param = findoption(options,'-p')
  ldap_port = string.atoi(param)
except: ldap_port

# Bind DN
try:
  dummy,ldap_binddn = findoption(options,'-D')
except: pass

# Password
try:
  dummy,ldap_passwd = findoption(options,'-w')
except: pass

# objectclass
try:
  dummy,ldap_entryobjectclass = findoption(options,'-o')
except: pass

if ldap_binddn and not ldap_passwd:
  sys.stderr.write('Passwort: ')
  ldap_passwd = string.strip(sys.stdin.readline())

ldap_matchmailonly = findoption(options,'-m')
ldap_createnew = findoption(options,'-n')
if findoption(options,'-r'):
  ldap_modifymode = ldap.MOD_REPLACE
else:
  ldap_modifymode = ldap.MOD_ADD
  
# Pfadname der SSLeay-Datenbank
try:
  dummy,openssl.db_filename = findoption(options,'-d')
except: pass
if not os.path.isfile(openssl.openssl.db_filename):
  PrintUsage('SSLeay database "%s" not found.' % openssl.openssl.db_filename)

# Verzeichnisname mit SSLeay-Zertfikaten im PEM-Format
try:
  dummy,openssl_certdir = findoption(options,'-c')
except: pass
if not os.path.isdir(openssl_certdir):
  PrintUsage('Directory "%s" with certificates not found.' % openssl_certdir)

# Pfadname des SSLeay-Programms
try:
  dummy,openssl_bin = findoption(options,'-s')
except: pass
if not os.path.isfile(openssl_bin):
  PrintUsage('SSLeay binary "%s" not found.' % openssl_bin)

if args:
  filter = {}
  for filterarg in args:
    try:
      filterattr,filtervalue = string.split(filterarg,'=')
    except ValueError:
      PrintUsage('Wrong filter format!')
    if filterattr in ldapattrname.keys():
      filter[string.strip(filterattr)]=string.strip(filtervalue)
    else:
      PrintUsage('Wrong filter attribute name!')
else:
  PrintUsage('No filter given.')

print 'Connect %s:%d' % (ldap_host,ldap_port)
l = ldap.open(ldap_host,ldap_port)
try:
  l.simple_bind_s(ldap_binddn,ldap_passwd)
except ldap.NO_SUCH_OBJECT:
  PrintErrorMsg('Unable to bind to host %s:%d as %s\nWrong password?' % (ldap_host,ldap_port,ldap_binddn))
  
allcertusers = openssl.GetEntriesbyDN(filter,0,1)

if not allcertusers:
  print 'No matching entries found in SSLeay database.'
  sys.exit(99)
  
print 'Found %d matching and valid entries in SSLeay database.' % len(allcertusers)

for certuser in allcertusers:

  openssldn = openssl.SplitDN(charset.iso2utf(charset.trans(certuser[openssl.DB_name],charset.t61,charset.iso)))
  # Eigentlich haben alle SSLeay-Zertifikate CN- und Email-Attribut,
  # aber wir fragen sicherheitshalber mal...
  if not openssldn.has_key('CN') or not openssldn.has_key('Email'):
    # CN- oder Email-Attribut fehlt => Fehlermeldung
    sys.stderr.write('Certificate line has no CN or Email-attribute!\n')
    
  # Erst einmal die passenden Eintraege suchen
  cn = openssldn['CN']
  email = openssldn['Email']
  ldapdnlist = []
  
  # Liste aller zu modifizierenden DNs auf LDAP-Server suchen
  # 1. Suche nach Eintraegen mit passenden cn- und mail-Attributen.
  print 'Searching LDAP for Common Name %s and mail address %s...' % (string.upper(cn),string.upper(email))
  ldapdnlist = l.search_s("",ldap.SCOPE_SUBTREE,'(&(cn=%s)(mail=%s))' % (string.upper(cn),string.upper(email)),None,1)
  
  # 2. Suche nach Eintraegen mit nur dem passenden mail-Attribut,
  #    falls Option -m angegeben
  if (not ldapdnlist) and ldap_matchmailonly:
      print 'Searching LDAP for mail address %s...' % (string.upper(email))
      ldapdnlist = l.search_s("",ldap.SCOPE_SUBTREE,'mail=%s' % string.upper(email),None,1)

  # 3. Wird ebenfalls kein Eintrag gefunden, so wird ein
  #    neuer Eintrag mit den Angaben aus dem Zertifikat angelegt,
  #    falls Option -c gesetzt
  if (not ldapdnlist) and ldap_createnew:
    dn = 'cn=%s,mail=%s' % (cn,email)
    entry = [('objectclass',['top','strongauthenticationuser',ldap_entryobjectclass]),
             ('usercertificate',' '),
             ('usercertificate;binary',' ')
            ]
    for attr in openssldn.keys():
      if ldapattrname.has_key(attr):
  	entry.append((ldapattrname[attr],[openssldn[attr]]))
      else:
  	entry.append((attr,[openssldn[attr]]))

    print 'Creating new entry "%s".' % (dn)
    l.add_s(dn,entry)
    ldapdnlist = [(dn,{})]

  # Zertifikat zu Eintraegen hinzufuegen
  for ldapdn in ldapdnlist:

    pemfilename = os.path.join(openssl_certdir,certuser[openssl.DB_serial]+'.pem')
    if not os.path.isfile(pemfilename):
      sys.stderr.write('Certificate %s file not found!\n' % pemfilename)

    # Erst mal PEM-Datei einlesen
    f = open(pemfilename,'r')
    certdatabuf = ''
    s = f.read()
    while s:
      certdatabuf = certdatabuf+s
      s = f.read()
    f.close()
    certdatabuf = string.translate(certdatabuf,string.maketrans("\n","$"),"\n")

    print 'Adding usercertificate %s (%d Bytes) to entry "%s".' % (pemfilename,len(certdatabuf),ldapdn[0])
    
    try:
      l.modify_s(ldapdn[0],[(ldap_modifymode,'usercertificate',certdatabuf)])
    except ldap.TYPE_OR_VALUE_EXISTS:
      PrintErrorMsg('Attribute "usercertificate;binary" already exists.\nYou might want to try option -r.',NoExit=1)
    except ldap.INSUFFICIENT_ACCESS:
      PrintErrorMsg('You are not allowed to modify entries.\nYou might want to use options -D and -w to bind with a specific DN and password.')

    # openssl x509 eine DER-Datei auf stdout schreiben lassen
    pipe = os.popen('%s x509 -outform der -in %s' %(openssl_bin,pemfilename))
    certdatabuf = ''
    s = pipe.read()
    while s:
      certdatabuf = certdatabuf+s
      s = pipe.read()
    pipe.close()
    
    print 'Adding usercertificate %s (%d Bytes) to entry "%s".' % (pemfilename,len(certdatabuf),ldapdn[0])
    
    try:
      l.modify_s(ldapdn[0],[(ldap_modifymode,'usercertificate;binary',certdatabuf)])
    except ldap.TYPE_OR_VALUE_EXISTS:
      PrintErrorMsg('Attribute "usercertificate;binary" already exists.\nYou might want to try option -r.',NoExit=1)
    except ldap.INSUFFICIENT_ACCESS:
      PrintErrorMsg('You are not allowed to modify entries.\nYou might want to use options -D and -w to bind with a specific DN and password.')
