#!/usr/bin/python

########################################################################
# ca-cycle.py
# (c) by Michael Stroeder, michael.stroeder@inka.de
########################################################################

Version = '0.5.3'

########################################################################
# This script is typically run by CRON or a similar task manager.
# It does several jobs (some not implemented yet):
# - Mark expired certificates in OpenSSL certificate database
# - Sort in new certificates and inform user via e-mail where to
#   download his certificate 
# - Spool certificate requests and certificate revocation requests 
# - Remove stale certificate requests from caPendCertReqDir
########################################################################

import sys, string, os, smtplib, time, pycacnf, openssl

from pycacnf import opensslcnf, pyca_section

MailRelay          = pyca_section.get('MailRelay','localhost')
nsGetCertUrl       = pyca_section.get('nsGetCertUrl','cgi-bin/get-cert.py')
caPendCertReqValid = 3600*string.atoi(pyca_section.get('caPendCertReqValid','0'))
OpenSSLExec = pyca_section.get('OpenSSLExec','/usr/local/ssl/bin/openssl')
caCertFormat = string.lower(pyca_section.get('caCertFormat','der'))

gmt = time.time()

######################################################################
# CA specific actions
######################################################################

ca_names = opensslcnf.sectionkeys.get('ca',[])

# Lists of processed files and dirs to avoid double-processing
processed_ca_databases = []
processed_pend_reqs_dirs = []
processed_new_certs_dirs = []

for ca_name in ca_names:

  sys.stdout.write('Processing certificate authority "%s"\n' % (ca_name))

  ca = opensslcnf.getcadata(ca_name)

  ######################################################################
  # Processing certificate database
  ######################################################################

  if not ca.database in processed_ca_databases:

    processed_ca_databases.append(ca.database)

    # Certificate database not processed up to now
    ca_db = openssl.db.OpenSSLcaDatabaseClass(ca.database)

    ####################################################################
    # Mark expired certificates in OpenSSL certificate database
    ####################################################################

    ca_db.Expire()

  ######################################################################
  # Remove stale certificate requests (SPKAC)
  ######################################################################

  if caPendCertReqValid and \
     (not ca.pend_reqs_dir in processed_pend_reqs_dirs) and \
     opensslcnf.isclientcert(ca_name):
  
    processed_pend_reqs_dirs.append(ca.pend_reqs_dir)
    # pend_certs_dir not processed up to now
    if os.path.isdir(ca.pend_reqs_dir):
      pendcertfilenames = os.listdir(ca.pend_reqs_dir)
    else:
      pendcertfilenames = []
      sys.stderr.write('Directory %s not found!\n' % (ca.pend_reqs_dir))

    stalecerts = 0

    for reqfilename in pendcertfilenames:

      reqpathname = os.path.join(ca.pend_reqs_dir,reqfilename)

      if gmt-caPendCertReqValid > os.path.getmtime(reqpathname):
        os.remove(reqpathname)
	stalecerts = stalecerts+1

    if stalecerts:
      sys.stdout.write('Removed %d stale certificate requests from %s.\n' % (stalecerts,ca.pend_reqs_dir))

  ######################################################################
  # Publish new client certificates and inform user via e-mail where to
  # download his certificate 
  ######################################################################

  if (not ca.new_certs_dir in processed_new_certs_dirs) and \
     opensslcnf.isclientcert(ca_name):

    processed_new_certs_dirs.append(ca.new_certs_dir)

    # new_certs_dir not processed up to now
    newcertfilenames = os.listdir(ca.new_certs_dir)
    
    if newcertfilenames:
      sys.stdout.write('Publish %d new certificates in %s.\n' % (len(newcertfilenames),ca.new_certs_dir))

    for certfilename in newcertfilenames:

      newcertpathname = os.path.join(ca.new_certs_dir,certfilename)
      cert = openssl.cert.X509CertificateClass(newcertpathname)
      
      if openssl.db.GetEntrybySerial(ca.database,cert.serial):

	# Mailbody
	mail_msg = """From: %s <%s>
To: %s <%s>
Subject: Your certificate %d

The certificate you requested has been issued.

You can retrieve your certificate here:

  %s%s/%s/user?%x

This will send your new certificate to you. Your browser will
probably open a dialogue asking you to make a backup
copy of your certificate including your private key.
This is highly recommended!

If you have further questions simply reply to this e-mail.
""" % \
(cert.issuer['CN'],cert.issuer['Email'],\
 cert.subject['CN'],cert.subject['Email'],\
 cert.serial,\
 ca.nsBaseUrl,nsGetCertUrl,ca_name,cert.serial)

	smtpconn=smtplib.SMTP(MailRelay)
	smtpconn.set_debuglevel(0)
	try:
	  smtpconn.sendmail(cert.issuer['Email'],[cert.subject['Email']],mail_msg)
	  sys.stderr.write('Sent e-mail to %s!\n' % (cert.subject['Email']))
	except:
	  sys.stderr.write('Unable to send an e-mail to %s!\n' % (cert.subject['Email']))
	smtpconn.quit()

	certpathname = os.path.join(ca.certs,'%s.%s' % (os.path.splitext(certfilename)[0],caCertFormat))
	rc = os.system('%s x509 -inform pem -outform "%s" -in "%s" -out "%s"' % ( \
                       OpenSSLExec,
		       caCertFormat,
		       newcertpathname,
		       certpathname
		       ))
	# everything ok?
	if not rc:
          # remove new cert in new_certs_dir
          os.remove(newcertpathname)
          # create symlink with hash value of cert as filename
  #        os.symlink(certpathname,os.path.join(ca.certs,'%s.%s' % (cert.hash,caCertFormat)))

      else:
        sys.stderr.write('Did not find cert serial %d in %s! Cert database up to date?\n' % (cert.serial,ca.database))

  ######################################################################
  # Spool certificate requests and certificate revocation requests
  # to system holding the private keys
  ######################################################################

