#!/usr/bin/python

########################################################################
# ca-make.py
# (c) by Michael Stroeder, michael.stroeder@inka.de
########################################################################

Version = '0.5.3'

########################################################################
# Make one or more certificate authorities by looking into openssl.cnf.
########################################################################
#
# This script creates if non-existent (in the order given below,
# does not overwrite existing files):
#
# Directory structure:
# dir			Where everything is kept
# certs			Where the issued certs are kept
# new_certs_dir		default place for new certs.
# crl_dir		Where the issued crl are kept
#
# Files:
# database		database index file.
# serial		The current serial number
#
# Certificate files:
# certificate		The CA certificate
# private_key		The private key of the CA
# crl			The initial CRL
########################################################################

import sys, string, os, stat, pycacnf, openssl

from pycacnf import opensslcnf, pyca_section

def filenotvalid(pathname):
  return not os.path.isfile(pathname) or os.stat(pathname)[stat.ST_SIZE]==0

def CheckedMakeDir(dirname,perms=0,uid=0,gid=0):
  if dirname:
    if os.path.exists(dirname):
      if not os.path.isdir(dirname):
	sys.stderr.write('%s already exists but is no directory.\n' % (dirname))
    else:
      sys.stdout.write('Creating %s\n' % (dirname))
      os.makedirs(dirname)
      if perms:
	sys.stdout.write('Changing permissions of %s to %o\n' % (dirname,perms))
	os.chmod(dirname,perms)
      if uid and gid:
	sys.stdout.write('Changing owner/group of %s to %d/%d\n' % (dirname,uid,gid))
	os.chown(dirname,uid,gid)



########################################################################
#                              Main
########################################################################

caCertFormat = string.lower(pyca_section.get('caCertFormat','der'))

OpenSSLExec = pyca_section.get('OpenSSLExec','/usr/local/ssl/bin/openssl')
if not os.path.isfile(OpenSSLExec):
  sys.stderr.write('Did not find OpenSSL executable %s.\n' % (OpenSSLExec))
  sys.exit(1)

ca_names = opensslcnf.sectionkeys.get('ca',[])

subca = []




sys.stdout.write("""
#############################################################
# Create directories and various files
#############################################################\n
""")

for ca_name in ca_names:

  ca = opensslcnf.getcadata(ca_name)
    
  # Create sub-directories
  CheckedMakeDir(ca.dir,perms=0755)
  CheckedMakeDir(ca.certs,perms=0755)
  CheckedMakeDir(ca.new_certs_dir,perms=0700)
  CheckedMakeDir(ca.crl_dir,perms=0755)
  CheckedMakeDir(ca.pend_reqs_dir)
  CheckedMakeDir(ca.new_reqs_dir)
  CheckedMakeDir(os.path.dirname(ca.certificate),perms=0755)
  CheckedMakeDir(os.path.dirname(ca.private_key),perms=0700)


  # FIX ME! The ownership/permissions of the spool
  # directories should be set here, e.g. like
  # d-wxrwx---   2 daemon   caadmin  newreqs/
  # d-wxrwx---   2 wwwrun   daemon   pendreqs/

  # database: database index file
  if not os.path.isfile(ca.database):
    sys.stdout.write('Creating database file %s\n' % (ca.database))
    file=open(ca.database,'w')
    file.write('')
    file.close()
  os.chmod(ca.database,0644)

  # serial: next serial number for issueing certificates
  if filenotvalid(ca.serial):
    sys.stdout.write('Creating serial file %s\n' % (ca.serial))
    file=open(ca.serial,'w')
    file.write('01\n')
    file.close()
  os.chmod(ca.serial,0600)



sys.stdout.write("""
#############################################################
# create CA certs or certificate requests
#############################################################\n
Give passwords for each CAs here.\n
""")

for ca_name in ca_names:

  ca = opensslcnf.getcadata(ca_name)
    
  if ca.signedby:
    subca.append(ca_name)

  if filenotvalid('%s-req' % ca.certificate) and filenotvalid(ca.private_key):
    sys.stdout.write('Creating certificate request %s with private key %s.\n' % (ca.certificate,ca.private_key))
    if not ca.ca_reqfile:
      ca.ca_reqfile = pycacnf.cnf_filename
    rc = os.system('%s req -config %s -new -outform pem -out %s-req -keyout %s' % \
                   (OpenSSLExec,ca.ca_reqfile,ca.certificate,ca.private_key))
#     rc = os.system('%s req -config %s -new -no-asn1-kludge -outform pem -out %s-req -keyout %s' % \
#                   (OpenSSLExec,ca.ca_reqfile,ca.certificate,ca.private_key))
    os.chmod(ca.private_key,0400)
    if rc:
      sys.stderr.write('Error %d creating CA cert request %s-req.\n' % (rc,ca.certificate))
    else:
      if not ca.signedby:
        sys.stdout.write('How long should this certificate be valid (default 30 days): ')
	days = string.strip(sys.stdin.readline())
	if not days:
	  days = 30
	rc = os.system('%s x509 -req -inform pem -in %s-req -outform pem -out %s -signkey %s -days %s -extfile %s' % \
                       (OpenSSLExec,ca.certificate,ca.certificate,ca.private_key,days,ca.ca_x509_extfile))
        if rc:
          sys.stderr.write('Error %d self-signing CA cert %s.\n' % (rc,ca.certificate))



if subca:
  sys.stdout.write("""
#############################################################
# Create certs of sub-CAs
#############################################################\n
Use passwords of parent CAs here.\n
""")

for ca_name in subca:

  subca = opensslcnf.getcadata(ca_name)
  ca = opensslcnf.getcadata(subca.signedby)

  if filenotvalid(subca.certificate):
    sys.stdout.write('Creating sub-CA certificate %s with issuer %s.\n' % (subca.certificate,ca.certificate))
    rc = os.system('%s x509 -req -inform pem -in %s-req -outform pem -out %s -CA %s -CAkey %s -CAserial %s -days %s -extfile %s' % \
                   (OpenSSLExec,subca.certificate,subca.certificate,ca.certificate,ca.private_key,ca.serial,ca.default_days,subca.ca_x509_extfile))
    if rc:
      sys.stderr.write('Error %d issueing CA cert %s.\n' % (rc,ca.certificate))



sys.stdout.write("""
#############################################################
# Create initial CRLs
#############################################################\n
Use passwords of each CA here.\n
""")

for ca_name in ca_names:

  ca = opensslcnf.getcadata(ca_name)

  if filenotvalid(ca.crl):
    if os.path.isfile(ca.certificate) and os.path.isfile(ca.private_key):
      sys.stdout.write('Creating initial CRL %s.\n' % (ca.crl))
      rc = os.system('%s ca -config %s -name %s -gencrl -out %s' % \
                     (OpenSSLExec,pycacnf.cnf_filename,ca.sectionname,ca.crl))
      if rc:
        sys.stderr.write('Error %d creating CRL %s.\n' % (rc,ca.crl))

    else:
      sys.stderr.write('CA cert %s or CA key %s missing.\n' % (ca.certificate,ca.private_key))



if caCertFormat!='pem':

  sys.stdout.write("""
#############################################################
# Convert ca-certificates and CRLs to caCertFormat
#############################################################\n
""")

  for ca_name in ca_names:

    ca = opensslcnf.getcadata(ca_name)
    # Convert CA cert
    if os.path.isfile(ca.certificate):
      ca.certificate_der = '%s.der' % os.path.splitext(ca.certificate)[0]
      sys.stdout.write('Converting CA cert %s to %s.\n' % (ca.certificate,ca.certificate_der))
      rc = os.system('%s x509 -inform pem -outform "%s" -in "%s" -out "%s"' % \
                     (OpenSSLExec,caCertFormat,ca.certificate,ca.certificate_der))
    else:
      sys.stderr.write('CA cert %s missing.\n' % (ca.certificate))
      
    # Convert CRL
    if os.path.isfile(ca.crl):
      ca_crl_der = '%s.der' % os.path.splitext(ca.crl)[0]
      sys.stdout.write('Converting CRL %s to %s.\n' % (ca.crl,ca_crl_der))
      rc = os.system('%s crl -inform pem -outform "%s" -in "%s" -out "%s"' % \
                     (OpenSSLExec,caCertFormat,ca.crl,ca_crl_der))
    else:
      sys.stderr.write('CRL %s missing.\n' % (ca.crl))



